"""
Provides a pythonic dict-like interface to local or remote leveldb.
In remote mode it allows access from multiple processes.
"""
import io
from pathlib import Path
from typing import Any, Iterable, Iterator, Optional, Union

from openleveldb.backend.connectorclient import LevelDBClient
from openleveldb.backend.connectorlocal import LevelDBLocal


class LevelDB:
    def __init__(
        self,
        db_path: Optional[Union[str, Path]],
        server_address: Optional[str] = None,
        dbconnector: Optional[Union[LevelDBLocal, LevelDBClient]] = None,
        read_only: bool = False,
    ) -> None:
        """
        Provide access to a leveldb database, it if does not exists one is created.

        Local databases do not support multiprocessing.
        It is possible to access a local db with:

            >>> db = LevelDB(db_path)

        Remote databases support multiprocessing. Once the a leveldb server is running,
        it is possible to access the remote db with:

            >>> # db = LevelDB(remote_db_path, server_address)

        :param db_path: the path in the filesystem to the database
        :param server_address: the address of the remote server
        :param read_only: if true the db can not be modified
        :param dbconnector: provide directly an existing dbconnector
        """
        self.db_path = db_path
        self.server_address = server_address
        self.dbconnector: Union[LevelDBClient, LevelDBLocal]
        self.read_only: bool = read_only

        if dbconnector is not None:
            self.dbconnector = dbconnector
        elif server_address is not None:
            self.dbconnector = LevelDBClient.get_instance(db_path, server_address)
        else:
            self.dbconnector = LevelDBLocal.get_instance(db_path=db_path)

    def prefixed_iter(
        self,
        prefixes: Optional[Union[str, Iterable[str]]] = None,
        starting_by: Optional[Union[str, Iterable[str]]] = None,
        include_key=True,
        include_value=True,
    ) -> Iterable:
        """
        Builds a custom iterator.

        The parameters ``include_key`` and  ``include_value`` define what should be
        yielded:

            >>> list(db)
            [('a1', 'value1'), ('b1', 'value2'), ('b2', 'value3'), ('c1', 'value4')]
            >>> list(db.prefixed_iter(include_key=False, include_value=False))
            [None, None, None, None]
            >>> list(db.prefixed_iter(include_key=True, include_value=False))
            ['a1', 'b1', 'b2', 'c1']
            >>> list(db.prefixed_iter(include_key=False, include_value=True))
            ['value1', 'value2', 'value3', 'value4']


        The ``prefixes`` and ``starting_by`` parameters have a similar meaning. They
        determine over which keys it should iterate. The difference is that
        ``starting by`` preserves the prefix in the returned key.
        The iterations stops when all the available keys with the given prefixes have
        been yielded

            >>> list(db)
            [('a1', 'value1'), ('b1', 'value2'), ('b2', 'value3'), ('c1', 'value4')]
            >>> list(db.prefixed_iter(prefixes=["b"]))
            [('1', 'value2'), ('2', 'value3')]
            >>> list(db.prefixed_iter(prefixes=["b", "1"]))
            [('', 'value2')]
            >>> list(db.prefixed_iter(starting_by="b"))
            [('b1', 'value2'), ('b2', 'value3')]
            >>> list(db.prefixed_iter(starting_by=["b", "1"]))
            [('b1', 'value2')]

        :param include_key: if False do not yield the keys
        :param include_value: if False do not yield the values

        :param prefixes: prefixes of the desired keys
            The prefixes  will  be removed from the keys returned

        :param starting_by: prefixes of the desired keys
            The prefixes  will  be preserved from the keys returned

        :return: the iterable over the keys and/or values
        """
        return self.dbconnector.prefixed_iter(
            prefixes=prefixes,
            starting_by=starting_by,
            include_key=include_key,
            include_value=include_value,
        )

    def prefixed_len(
        self,
        prefixes: Optional[Union[str, Iterable[str]]] = None,
        starting_by: Optional[str] = None,
    ) -> int:
        """
        Utility function to compute the number of keys with a given prefix,
        see :py:meth:~database.LevelDB.prefixed_iter for more details.

            >>> list(db)
            [('a1', 'value1'), ('b1', 'value2'), ('b2', 'value3'), ('c1', 'value4')]
            >>> db.prefixed_len(prefixes=["b"])
            2
            >>> db.prefixed_len(prefixes=["b", "1"])
            1

        :param prefixes: prefixes of the desired keys
            The prefixes  will  be removed from the keys returned

        :param starting_by: prefixes of the desired keys
            The prefixes  will  be preserved from the keys returned

        :return: the number of matching keys
        """
        return self.dbconnector.prefixed_len(
            prefixes=prefixes, starting_by=starting_by,
        )

    def __iter__(self) -> Iterator:
        """
        Iterator over (key, value) sorted by key

            >>> list(db)
            [('a1', 'value1'), ('b1', 'value2'), ('b2', 'value3'), ('c1', 'value4')]

        :returns: the iterator over the items
        """
        return iter(self.dbconnector)

    def __len__(self) -> int:
        """
        Computes the number of element in the database.

            >>> list(db)
            [('a1', 'value1'), ('b1', 'value2'), ('b2', 'value3'), ('c1', 'value4')]
            >>> len(db)
            4

        :returns: number of elements in the database
        """
        return len(self.dbconnector)

    def __setitem__(self, key: Union[str, Iterable[str]], value: Any) -> None:
        """
        Store the couple (key, value) in leveldb.
        The key and the value are automatically encoded together with the obj
        type information, in order to be able to automate the decoding.

            >>> import numpy as np
            >>> db["array"] = np.array([1, 2, 3], dtype=np.int8)
            >>> db["array"]
            array([1, 2, 3], dtype=int8)
            >>> del db["array"]


        The key may be one or more strings to specify prefixes.
        The last element is always the key:

            >>> db["prefix1", "prefix2", "key"] = "myvalue"
            >>> db["prefix1", "prefix2", "key"]
            'myvalue'
            >>> del db["prefix1", "prefix2", "key"]

        :param key: one or more strings to specify prefixes
        :returns: the value associated to the key in leveldb
        """
        if self.read_only:
            raise io.UnsupportedOperation("not writable")
        self.dbconnector[key] = value

    def __getitem__(
        self, key: Union[str, Iterable[Union[str, Ellipsis.__class__]]]
    ) -> Any:
        """
        Retrieve the couple (key, value) from leveldb.

        >>> db["a1"]
        'value1'
        >>> db["a", "1"]
        'value1'

        The value is automatically decoded into its original type. The value must
        have been stored with :py:~:py:meth:~database.LevelDB.__setitem__

            >>> import numpy as np
            >>> db["array"] = np.array([1, 2, 3], dtype=np.int8)
            >>> db["array"]
            array([1, 2, 3], dtype=int8)
            >>> del db["array"]

        The key may be one or more strings, to specify prefixes.
        The last element is always the key:

            >>> db["prefix1", "prefix2", "key"] = "myvalue"
            >>> db["prefix1", "prefix2", "key"]
            'myvalue'
            >>> del db["prefix1", "prefix2", "key"]

        It is possible to retrieve a stateful instance of :py:class:~database.LevelDB
        that accounts for prefixes using Ellipsis as key:

            >>> list(db)
            [('a1', 'value1'), ('b1', 'value2'), ('b2', 'value3'), ('c1', 'value4')]
            >>> db_b = db['b', ...]
            >>> db_b["1"]
            'value2'
            >>> list(db_b)
            [('1', 'value2'), ('2', 'value3')]
            >>> list(db["c", ...])
            [('1', 'value4')]

        :param key: one or more strings to specify prefixes. It's possible to specify
            sub-db using the Ellipsis as key.
        :returns: the value associated to the key in leveldb or a sub-db.
        """
        out = self.dbconnector[key]
        return (
            out
            if not isinstance(out, type(self.dbconnector))
            else LevelDB(
                db_path=self.db_path,
                server_address=self.server_address,
                dbconnector=out,
                read_only=self.read_only,
            )
        )

    def __delitem__(self, key: Union[str, Iterable[str]]) -> None:
        """
        Delete the couple (key, value) from leveldb.

        The key may be one or more strings, to specify prefixes.
        The last element is always the key:


            >>> db["prefix", "key"] = "value"
            >>> db["prefix", "key"]
            'value'
            >>> db["prefixkey"]
            'value'
            >>> del db["prefix", "key"]
            >>> print(db["prefix", "key"])
            None
            >>> print(db["prefixkey"])
            None

        :param key: one or more strings to specify prefixes.
        """
        if self.read_only:
            raise io.UnsupportedOperation("not writable")
        del self.dbconnector[key]

    def __repr__(self) -> str:
        name = self.__class__.__name__
        path = self.db_path
        ip = f"'{self.server_address}'" if self.server_address is not None else None
        return (
            f"{name}(db_path='{path}', "
            f"server_address={ip}, "
            f"dbconnector={self.dbconnector})"
        )

    def close(self) -> None:
        """
        Close the database
        """
        self.dbconnector.close()


if __name__ == "__main__":
    import doctest
    import tempfile

    import numpy as np

    db_path = tempfile.mkdtemp()
    db = LevelDB(db_path=db_path)

    db["a", "1"] = "value1"
    db["b", "1"] = "value2"
    db["b", "2"] = "value3"
    db["c", "1"] = "value4"

    doctest.testmod(extraglobs={"tmpfile": db_path, "db": db})



"""
Automatic conversion to and from bytes
"""
import base64
import collections
import functools
import json
import typing
from enum import Enum
from io import BytesIO
from typing import Any, Callable, Iterable, Union

import numpy as np
import rapidjson


class Serializer:
    @staticmethod
    def ndarray_frombytes(blob: bytes) -> np.ndarray:
        """
        Convert the bytes representation of the array back into a numpy array.

        Use BytesIO to convert the binary version of the array back into a numpy array.

        :param blob: BLOB containing a NumPy array
        :returns: One steaming hot NumPy array
        """
        out = BytesIO(blob)
        out.seek(0)

        return np.load(out)

    @staticmethod
    def ndarray_tobytes(array: np.ndarray) -> bytes:
        """
        Convert the numpy array into bytes

        Use the numpy.save function to save a binary version of the array,
        and BytesIO to catch the stream of data and convert it into a BLOB.

        :param array: NumPy array to turn into bytes
        :returns: NumPy array as BLOB
        """
        out = BytesIO()
        np.save(out, array)
        out.seek(0)

        return out.read()

    @staticmethod
    def bytes_to_string(blob: bytes) -> str:
        """
        Convert a numpy array into an ascii representation

        Using the numpy.save function to save a binary version of the array,
        and BytesIO to catch the stream of data and convert it into a BLOB.
        Then convert the BLOB in a base64 ascii representation.

        :param obj: NumPy array to turn into bytes
        :returns: NumPy array as an ascii string
        """
        byte_string = base64.b64encode(blob)
        ndarray_string = byte_string.decode("ascii")
        return ndarray_string

    @staticmethod
    def string_to_bytes(obj: str) -> bytes:
        """
        Convert the ascii representation of the array back into a numpy array.

        :param obj: ascii string containing a NumPy array
        :returns: One steaming hot NumPy array
        """
        blob = base64.b64decode(obj)
        return blob


class DecodeError(BaseException):
    pass


class DecodeType(Enum):
    """
    Enum that establish how each object should be serialized into bytes and de-serialized.
    Perform automatic decode adding an identifier to the byte representation of the objects

    Changing this class may yield compatibility issues with old data.
    """

    @staticmethod
    def _id_length() -> int:
        """
        The expected length of the identifier that will be concatenated to the bytes representation.
        Changing this return value will break old data

        :return: the identifier length
        """
        return 1

    def __new__(
        cls,
        id: bytes,
        encode_fun: Callable[[Any], bytes],
        decode_fun: Callable[[bytes], Any],
    ) -> "DecodeType":
        """
        Custom constructor for enum constants.
        The constants are indexed only by the identifier, but the have other attributes

        :param id: the identifier that will be concatenated to the byte representation
        :param encode_fun: the function to use to convert objects into bytes
        :param decode_fun: the function to use to convert bytes into objects
        """
        if len(id) != cls._id_length():
            raise TypeError(f"unsupported {len(id)}-byte identifier '{id}'")

        obj = object.__new__(cls)
        obj._value_ = obj.identifier = id
        obj.pure_encode_fun = encode_fun
        obj.pure_decode_fun = decode_fun
        return obj

    def encode(self, obj: Any) -> bytes:
        """
        Encode method, available for every constants.
        Encodes the object into bytes and concatenates the identifier

        :param obj: the object to encode
        :return: the encoded object
        """
        return self.identifier + self.pure_encode_fun(obj)

    @staticmethod
    def decode(blob: bytes) -> Any:
        """
        Decode method, available for every constants.
        Decodes the bytes blob into the object, uses the identifier contained in the blob to select the right decoder.

        :param blob: the blob to decode
        :return: the decoded object
        """
        idlen = DecodeType._id_length()
        enc, obj = blob[:idlen], blob[idlen:]
        try:
            return DecodeType(enc).pure_decode_fun(obj)
        except json.decoder.JSONDecodeError as ex:
            raise ex
        except ValueError as ex:
            raise DecodeError(f"missing DecodeType identifier in bytes blob") from ex

    BYTES = (b"b", lambda x: x, lambda x: x)
    INT = (
        b"i",
        functools.partial(int.to_bytes, length=16, byteorder="big", signed=True),
        functools.partial(int.from_bytes, byteorder="big", signed=True),
    )
    STR = (
        b"s",
        functools.partial(str.encode, encoding="utf-8"),
        functools.partial(bytes.decode, encoding="utf-8"),
    )
    JSON = (
        b"j",
        lambda x: str.encode(rapidjson.dumps(x), encoding="utf-8"),
        lambda x: rapidjson.loads(bytes.decode(x, encoding="utf-8")),
    )
    NUMPY = (
        b"n",
        Serializer.ndarray_tobytes,
        Serializer.ndarray_frombytes,
    )


# todo: use singledispatchmethod and move inside the enum in python 3.8
@functools.singledispatch
def encode(obj: Any) -> bytes:
    """
    Utility method to select the correct DecodeType to perform the encoding, dispatching based on the obj type
    Encode the object into bytes, the encoding may change depending on the obj type

    It supports json serializable, string, int and np.ndarray

    :param obj: the object to encode
    :return: the encoded object
    """
    return DecodeType.JSON.encode(obj)


@encode.register
def _(obj: bytes) -> bytes:
    return DecodeType.BYTES.encode(obj)


@encode.register
def _(obj: int) -> bytes:
    return DecodeType.INT.encode(obj)


@encode.register
def _(obj: str) -> bytes:
    return DecodeType.STR.encode(obj)


@encode.register
def _(obj: np.ndarray) -> bytes:
    return DecodeType.NUMPY.encode(obj)


@functools.singledispatch
def decode(blob: Any) -> Any:
    """
    Utility method to expose DecodeType.decode

    Decode the bytes back into an object. The blob is expected to contain a type identifier in the first 4 bytes.
    The decoding may change depending on blob's identifier

    :param blob: the blob to decode
    :return: the decoded object
    """
    return blob


@decode.register
def _(blob: bytes) -> Any:
    return DecodeType.decode(blob)


def normalize_strings(
    key_encoder: Callable[[str], bytes], strings: Union[str, typing.Sequence[str]]
) -> Iterable[bytes]:
    """
    Normalize an iterable of strings to the correspondent iterable of
    bytes-encoded strings. It always returns an Iterable

    :param strings: the string or iterable of strings to encode
    :param key_encoder: the encoder to use to encode the strings
    :returns: the encoded strings
    """
    if strings is None:
        yield None

    if strings is Ellipsis:
        yield strings

    if isinstance(strings, bytes):
        raise TypeError(f"str prefix or key expected, got {type(strings).__name__}")

    if isinstance(strings, str):
        strings = [strings]

    if not isinstance(strings, collections.abc.Sequence):
        raise TypeError(f"str prefix or key expected, got {type(strings).__name__}")

    num_strings = len(strings)

    for i, s in enumerate(strings):
        if i == num_strings - 1 and s is Ellipsis:
            yield s
        elif not isinstance(s, str):
            raise TypeError(f"str prefix or key expected, got {type(s).__name__}")
        else:
            yield key_encoder(s)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Class that exposes the leveldb through REST API, with automatic serialization
to bytes and deserialization from bytes provided by the serialization module
"""
import http
import os
import pickle
import sys
from multiprocessing import Process
from pathlib import Path
from typing import Iterable, Tuple, Union

import plyvel
from flask import Flask, Response, g, request
from openleveldb.backend import serializer
from openleveldb.backend.connectorcommon import get_prefixed_db
from openleveldb.backend.serializer import DecodeType

app = Flask(__name__)


def get_db(dbpath: Union[str, Path]) -> plyvel.DB:
    dbpath: Path = Path(dbpath)

    if not hasattr(g, "dbs"):
        g.dbs = {}

    if dbpath not in g.dbs:
        g.dbs[dbpath] = plyvel.DB(
            dbpath.expanduser().absolute().as_posix(), create_if_missing=True
        )

    return g.dbs[dbpath]


def _parse_and_get_prefixed_db() -> plyvel.DB:
    dbpath = request.args.get("dbpath")
    prefixes = request.args.getlist("prefixes")
    prefixes = (
        serializer.normalize_strings(DecodeType.STR.pure_encode_fun, prefixes)
        if prefixes is not None
        else ()
    )
    return get_prefixed_db(get_db(dbpath), prefixes)


@app.teardown_appcontext
def close_db(error) -> None:
    """Closes the database again at the end of the request."""
    if hasattr(g, "dbs"):
        for x, y in g.dbs.items():
            if hasattr(y, "close"):
                y.close()


@app.route("/iterator", methods=["GET"])
def iterator() -> Iterable[Union[bytes, Tuple[bytes, bytes]]]:
    db = _parse_and_get_prefixed_db()

    starting_by = request.args.get("starting_by")
    starting_by = b"".join(
        serializer.normalize_strings(DecodeType.STR.pure_encode_fun, starting_by)
        if starting_by is not None
        else ()
    )

    include_key = request.args.get("include_key") == "True"
    include_value = request.args.get("include_value") == "True"

    out = pickle.dumps(
        list(
            db.iterator(
                prefix=starting_by, include_key=include_key, include_value=include_value
            )
        )
    )
    return Response(out, content_type="application/octet-stream")


@app.route("/dblen", methods=["GET"])
def dblen() -> str:
    db = _parse_and_get_prefixed_db()

    starting_by = request.args.get("starting_by")
    starting_by = b"".join(
        serializer.normalize_strings(DecodeType.STR.pure_encode_fun, starting_by)
        if starting_by is not None
        else ()
    )
    out = serializer.encode(
        sum(
            1
            for _ in db.iterator(
                include_key=True, include_value=False, prefix=starting_by
            )
        )
    )
    return Response(out, content_type="application/octet-stream")


@app.route("/setitem", methods=["POST"])
def setitem() -> Response:
    db = _parse_and_get_prefixed_db()
    key = request.args.get("key")
    value = request.get_data()
    keybytes = DecodeType.STR.pure_encode_fun(key)

    db.put(keybytes, value)

    return Response(key, content_type="text")


@app.route("/getitem", methods=["GET"])
def getitem() -> Response:
    db = _parse_and_get_prefixed_db()
    key = request.args.get("key")
    keybytes = DecodeType.STR.pure_encode_fun(key)
    out = db.get(keybytes, default=b"")
    return Response(out, content_type="application/octet-stream")


@app.route("/delitem", methods=["DELETE"])
def delitem() -> (str, http.HTTPStatus):
    db = _parse_and_get_prefixed_db()

    key = request.args.get("key")
    keybytes = DecodeType.STR.pure_encode_fun(key)
    db.delete(keybytes)

    return Response(key, content_type="text")


@app.route("/repr", methods=["GET"])
def repr() -> str:
    db = _parse_and_get_prefixed_db()

    dbpath = request.args.get("dbpath")
    classname = request.args.get("classname")
    innerdb = f"{db}"

    dbrepr = f"{classname}(path='{dbpath}', db={innerdb})"
    return Response(dbrepr, content_type="text")


def dummy_server(port: Union[int, str]) -> Process:
    port = int(port)

    def runflask() -> None:
        sys.stdout = open(os.devnull, "w")
        sys.stderr = open(os.devnull, "w")
        app.run(port=port)

    dummy_server = Process(target=runflask)
    dummy_server.start()
    return dummy_server


if __name__ == "__main__":
    pass



"""
Class that wraps the LevelDB plyvel to support dict-like behaviour,
with automatic serialization to bytes and deserialization from bytes provided
by the serialization module
"""
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Union

import plyvel
from openleveldb.backend import serializer
from openleveldb.backend.connectorcommon import get_prefixed_db
from openleveldb.backend.serializer import normalize_strings


class LevelDBLocal:

    _instances = {}

    @staticmethod
    def get_instance(db_path: Union[str, Path]) -> "LevelDBLocal":
        db_path = Path(db_path)
        if db_path not in LevelDBLocal._instances:
            LevelDBLocal._instances[db_path] = LevelDBLocal(db_path)
        return LevelDBLocal._instances[db_path]

    def __init__(
        self,
        db_path: Optional[Union[str, Path]],
        db: plyvel.DB = None,
        key_encoder: Callable[[str], bytes] = serializer.DecodeType.STR.pure_encode_fun,
        key_decoder: Callable[[bytes], str] = serializer.DecodeType.STR.pure_decode_fun,
        value_encoder: Callable[[Any], bytes] = serializer.encode,
        value_decoder: Callable[[bytes], Any] = serializer.decode,
    ) -> None:
        self.key_encoder, self.key_decoder = key_encoder, key_decoder
        self.value_encoder, self.value_decoder = value_encoder, value_decoder

        self.db_path: Path = Path(db_path) if db_path is not None else None
        if self.db_path in LevelDBLocal._instances:
            raise RuntimeError(
                f"{LevelDBLocal._instances[self.db_path]} already exists."
            )

        if db is not None:
            self.db = db
            return

        LevelDBLocal._instances[
            self.db_path
        ] = None  # Ensure the singleton pattern is enforced

        self.db = plyvel.DB(
            self.db_path.expanduser().absolute().as_posix(), create_if_missing=True
        )
        self.iter_params: dict = {}
        self.iter_prefixes: Optional[str, Iterable[str]] = None
        self.iter_starting_by: Optional[str] = None

    def prefixed_db(self, prefixes: Iterable[bytes]) -> "LevelDBLocal":
        return LevelDBLocal(
            db_path=None,
            db=get_prefixed_db(self.db, prefixes) if prefixes is not None else self.db,
        )

    # todo: must be tested
    def prefixed_iter(
        self,
        prefixes: Optional[Union[str, Iterable[str]]] = None,
        starting_by: Optional[str] = None,
        include_key=True,
        include_value=True,
    ) -> Iterable:
        prefixes = (
            normalize_strings(self.key_encoder, prefixes)
            if prefixes is not None
            else ()
        )
        starting_by = b"".join(
            serializer.normalize_strings(self.key_encoder, starting_by)
            if starting_by is not None
            else ()
        )
        subdb = self.prefixed_db(prefixes=prefixes)
        for x in subdb.db.iterator(
            prefix=starting_by, include_key=include_key, include_value=include_value
        ):
            if isinstance(x, bytes):
                if include_key:
                    yield self.key_decoder(x)
                else:
                    yield self.value_decoder(x)

            else:
                try:
                    key, value = x
                    yield self.key_decoder(key), self.value_decoder(value)
                except TypeError:
                    yield None

    def prefixed_len(
        self,
        prefixes: Optional[Union[str, Iterable[str]]] = None,
        starting_by: Optional[str] = None,
    ) -> int:
        return sum(
            1
            for _ in self.prefixed_iter(
                prefixes=prefixes,
                starting_by=starting_by,
                include_key=True,
                include_value=False,
            )
        )

    def __iter__(self) -> Iterable:
        return self.prefixed_iter()

    def __len__(self) -> int:
        return sum(1 for _ in self.prefixed_iter(include_key=True, include_value=False))

    def __setitem__(self, key: Union[str, Iterable[str]], value: Any) -> None:
        *prefixes, key = normalize_strings(self.key_encoder, key)

        if key is Ellipsis:
            raise TypeError(f"str prefix or key expected, got {type(key).__name__}")

        ldb = self.prefixed_db(prefixes)
        ldb.db.put(key, self.value_encoder(value))

    def __getitem__(
        self, key: Union[str, Iterable[Union[str, Ellipsis.__class__]]]
    ) -> Any:
        *prefixes, key = normalize_strings(self.key_encoder, key)
        ldb = self.prefixed_db(prefixes)
        if key is Ellipsis:
            return ldb
        out = ldb.db.get(key)
        return self.value_decoder(out) if out else None

    def __delitem__(self, key: Union[str, Iterable[str]]) -> None:
        *prefixes, key = normalize_strings(self.key_encoder, key)
        ldb = self.prefixed_db(prefixes)
        ldb.db.delete(key)

    def __repr__(self) -> str:
        innerdb = f"{self.db}"
        return f"{self.__class__.__name__}(path='{self.db_path}', db={innerdb})"

    def close(self) -> None:
        self.db.close()


if __name__ == "__main__":
    import doctest
    import tempfile

    tmpfile = tempfile.mkdtemp()
    doctest.testmod(extraglobs={"tmpfile": tmpfile})



import os

import dotenv


def load_envs(env_file: str = "system.env") -> None:
    """
    Load all the environment variables defined in the `env_file`.
    This is equivalent to `. env_file` in bash.

    It is possible to define all the system specific variables in the `env_file`.

    :param env_file: the file that defines the environment variables to use
    """
    if not os.path.isfile(env_file):
        raise FileNotFoundError(f"no such config file: {env_file}")
    dotenv.load_dotenv(dotenv_path=env_file, override=True)


def get_env(env_name: str) -> str:
    """
    Read an environment variable.
    Raises errors if it is not defined or empty.

    :param env_name: the name of the environment variable
    :return: the value of the environment variable
    """
    if env_name not in os.environ:
        raise NameError(f"name {env_name} is not defined")
    env_value: str = os.environ[env_name]
    if not env_value:
        raise ValueError(f"{env_name} is defined but empty")
    return env_value



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['openleveldb', 'openleveldb.backend']

package_data = \
{'': ['*']}

install_requires = \
['flask>=1.1.1,<2.0.0',
 'numpy>=1.18.1,<2.0.0',
 'plyvel>=1.1.0,<2.0.0',
 'pytest>=5.4.1,<6.0.0',
 'python-dotenv>=0.10.3,<0.11.0',
 'python-rapidjson>=0.9.1,<0.10.0',
 'requests>=2.22.0,<3.0.0']

setup_kwargs = {
    'name': 'openleveldb',
    'version': '0.1.4',
    'description': '',
    'long_description': '============================\nA pythonic leveldb wrapper\n============================\n|pypi| |docs| |license| |black|\n\n.. inclusion-marker-do-not-remove\n\n\n`Openleveldb <https://openleveldb.readthedocs.io/en/latest/index.html#>`_ is a small pythonic wrapper around Plyvel_\n\n\nFeatures\n========\n\n\nTransparent object store\n------------------------\n\nIt works with python objects:\n\n- Automatically **encodes objects** into bytes when saving to leveldb\n- Automatically **decodes bytes** into their original type when retrieving objects from leveldb\n\nSupported types include:\n\n- ``int``\n- ``str``\n- ``numpy.ndarray``\n- Anything that is serializable by orjson_\n\n>>> db[\'key\'] = {\'key\': [1, 2, 3]}\n>>> db[\'key\']\n{\'key\': [1, 2, 3]}\n\n\nPython dict-like protocol\n-------------------------\n\nIt offers dict-like interface to LevelDB_\n\n\n>>> db["prefix", "key"] = np.array([1, 2, 3], dtype=np.int8)\n>>> db["prefix", "key"]\narray([1, 2, 3], dtype=int8)\n\n>>> db = db["prefix", ...]\n>>> db["key"]\narray([1, 2, 3], dtype=int8)\n\n\n\nString-only keys\n----------------\n\nThe only possible type for the keys is ``str``.\nIt avoids several problems when working with prefixes.\n\n\n\nMultiprocessing support\n-----------------------\n\nExperimental **multiprocessing** support using a background flask server,\nexposing the same API of a direct connection::\n\n    db = LevelDB(db_path="path_to_db", server_address="http://127.0.0.1:5000")\n\n\n.. _Plyvel: https://github.com/wbolster/plyvel\n.. _LevelDB: http://code.google.com/p/leveldb/\n.. _orjson: https://github.com/ijl/orjson\n\n\n.. |docs| image:: https://readthedocs.org/projects/openleveldb/badge/?version=latest\n    :target: https://openleveldb.readthedocs.io/en/latest/?badge=latest\n    :alt: Documentation Status\n\n.. |license| image:: https://img.shields.io/github/license/lucmos/openleveldb\n    :target: https://github.com/lucmos/openleveldb/blob/master/LICENSE\n    :alt: Openleveldb license\n    \n.. |pypi| image:: https://img.shields.io/pypi/v/openleveldb\n    :target: https://pypi.org/project/openleveldb/\n    :alt: Openleveldb repo\n\n.. |black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/psf/black\n    :alt: Black syntax\n',
    'author': 'Luca Moschella',
    'author_email': 'luca.moschella94@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lucmos/openleveldb',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.5,<4.0.0',
}


setup(**setup_kwargs)



import random
import uuid
from typing import Iterable

import plyvel


class RandomUuidGenerator:
    """Generates a sequence of pseudo-random UUIDs.

    Given the same seed, it will generate the same sequence.
    """

    def __init__(self, seed: int) -> None:
        self.rng = random.Random(seed)

    def gen_uuid(self) -> uuid.UUID:
        return uuid.UUID(
            bytes=bytes(self.rng.getrandbits(8) for _ in range(16)), version=4
        )


a = RandomUuidGenerator(0)

BYTES_DELIMITER = a.gen_uuid().bytes


def get_prefixed_db(db: plyvel.DB, prefixes: Iterable[bytes]) -> plyvel.DB:
    """
    Apply all the prefixes (last one included) to obtain the desired prefixed database

    :param db: the initial database
    :param prefixes: the prefix or the iterable of prefixes to apply
    :returns: the prefixed database
    """

    for prefix in prefixes:
        if prefix is Ellipsis:
            raise TypeError(f"str prefix or key expected, got {type(prefix).__name__}")
        db = db.prefixed_db(prefix)

    return db



"""
Class that communicates with the server module to support dict-like behaviour,
with automatic serialization to bytes and deserialization from bytes provided
by the serialization module
"""

import pickle
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Union

import requests

from flask import Response
from openleveldb.backend.serializer import DecodeType, decode, encode, normalize_strings


class LevelDBClient:

    _instances = {}

    @staticmethod
    def get_instance(
        db_path: Union[str, Path], server_address: str,
    ) -> "LevelDBClient":
        db_path = Path(db_path)
        if db_path not in LevelDBClient._instances:
            LevelDBClient._instances[(server_address, db_path)] = LevelDBClient(
                server_address, db_path
            )
        return LevelDBClient._instances[(server_address, db_path)]

    def __init__(
        self,
        server_address: str,
        db_path: Union[str, Path],
        value_encoder: Callable[[Any], bytes] = encode,
        value_decoder: Callable[[bytes], Any] = decode,
    ) -> None:
        self.value_encoder, self.value_decoder = value_encoder, value_decoder

        self.server_address = server_address
        self.db_path = db_path

    def _prefixed_db(self, prefixes: Iterable[str]) -> str:
        """
        Apply all the prefixes (last one included) to obtain the desired prefixed database

        :param prefixes: the prefix or the iterable of prefixes to apply
        :returns: the prefixed database
        """
        res = requests.get(
            url=self.server_address + "/get_prefixed_db_path",
            params={"prefixes": prefixes, "dbpath": self.db_path},
        )
        return res.text

    def prefixed_iter(
        self,
        prefixes: Optional[Union[str, Iterable[str]]] = None,
        starting_by: Optional[str] = None,
        include_key=True,
        include_value=True,
    ) -> Iterable:
        res = requests.get(
            url=self.server_address + "/iterator",
            params={
                "dbpath": self.db_path,
                "include_key": include_key,
                "include_value": include_value,
                "starting_by": starting_by,
                "prefixes": prefixes,
            },
        )
        for x in pickle.loads(res.content):
            if isinstance(x, bytes):
                if include_key:
                    yield DecodeType.STR.pure_decode_fun(x)
                else:
                    yield self.value_decoder(x)

            else:
                try:
                    key, value = x
                    yield DecodeType.STR.pure_decode_fun(key), self.value_decoder(value)
                except TypeError:
                    yield None

    def __iter__(self) -> Iterable:
        return self.prefixed_iter([], None, True, True)

    def prefixed_len(
        self,
        prefixes: Optional[Union[str, Iterable[str]]] = None,
        starting_by: Optional[str] = None,
    ) -> int:
        res = requests.get(
            url=self.server_address + "/dblen",
            params={
                "dbpath": self.db_path,
                "prefixes": prefixes,
                "starting_by": starting_by,
            },
        )
        return decode(res.content)

    def __len__(self) -> int:
        res = requests.get(
            url=self.server_address + "/dblen",
            params={"dbpath": self.db_path, "prefixes": None, "starting_by": None,},
        )
        return decode(res.content)

    def __setitem__(self, key: Union[str, Iterable[str]], value: Any) -> Response:
        *prefixes, key = normalize_strings(lambda x: x, key)

        if key is Ellipsis:
            raise TypeError(f"str prefix or key expected, got {type(key).__name__}")

        res = requests.post(
            url=self.server_address + "/setitem",
            data=encode(value),
            headers={"Content-Type": "application/octet-stream"},
            params={"dbpath": self.db_path, "key": key, "prefixes": prefixes},
        )
        return res

    def __getitem__(
        self, key: Union[str, Iterable[Union[str, Ellipsis.__class__]]]
    ) -> Any:
        *prefixes, key = normalize_strings(lambda x: x, key)

        if key is Ellipsis:
            raise NotImplementedError

        res = requests.get(
            url=self.server_address + "/getitem",
            params={"dbpath": self.db_path, "key": key, "prefixes": prefixes},
        )

        return self.value_decoder(res.content) if res.content else None

    def __delitem__(self, key: Union[str, Iterable[str]]) -> Response:
        *prefixes, key = normalize_strings(lambda x: x, key)

        res = requests.delete(
            url=self.server_address + "/delitem",
            params={"dbpath": self.db_path, "key": key, "prefixes": prefixes},
        )
        return res

    def __repr__(self) -> str:
        res = requests.get(
            url=self.server_address + "/repr",
            params={"dbpath": self.db_path, "classname": self.__class__.__name__},
        )
        return res.text

    def close(self) -> None:
        pass


if __name__ == "__main__":
    pass
